unit uTiny;

interface

uses
  Classes;

type
  TOnCompilerMessage = procedure(const msg : string; var stop : boolean) of object;
  TTinyComp = class
  private
    fNBCSrc : TStrings;
    fMessages : TStrings;
    fMS : TMemoryStream;
    fTempChar : Char;
    fCCSet : boolean;
    fIncludeDirs: TStrings;
    fCurFile: string;
    fOnCompMSg: TOnCompilerMessage;
    fDirLine : string;
    fCurrentLine : string;
    fAPIFunctions : TStrings;
    procedure GetCharX;
    procedure GetChar;
    procedure Init;
    procedure Prog;
    procedure SkipCommentBlock;
    procedure SkipCommentLine;
    procedure SkipDirectiveLine;
    procedure SkipWhite;
    procedure GetName;
    procedure GetNum;
    procedure GetHexNum;
    procedure GetCharLit;
    procedure GetOp;
    procedure Next;
    procedure MatchString(x: string);
    procedure Semi;
    procedure Factor;
    procedure Divide;
    procedure Multiply;
    procedure Term;
    procedure Add;
    procedure Equal;
    procedure Expression;
    procedure LessOrEqual;
    procedure NotEqual;
    procedure Subtract;
    procedure BoolExpression;
    procedure Greater;
    procedure LeftShift;
    procedure Less;
    procedure NEqual;
    procedure Relation;
    procedure RightShift;
    procedure BoolOr;
    procedure BoolTerm;
    procedure BoolXor;
    procedure NotFactor;
    procedure StringConst;
    procedure DoCall(procname: string);
    procedure DoCallAPIFunc(procname: string);
    procedure Assignment;
    procedure Block(const lend : string = ''; const lstart : string = '');
    procedure BlockStatements(const lend : string = ''; const lstart : string = '');
    procedure DoFor;
    procedure DoIf(const lend, lstart : string);
    procedure DoRead;
    procedure DoWhile;
    procedure DoDoWhile;
    procedure DoRepeat;
    procedure DoAsm;
    procedure DoSwitch;
    procedure DoWrite;
    procedure ReadVar;
    procedure Statement(const lend, lstart : string);
    procedure ProcessDirectives;
    procedure HandlePoundLine;
    procedure Alloc(dt: char);
    function GetVariableType(vt: char; bUnsigned: boolean): char;
    procedure TopDecls;
    procedure IncLineNumber;
    procedure AddLocal(name: string; dt: char);
    procedure AllocLocal(const sub: string; dt: char);
    procedure DoLocals(const sub: string);
    function FormalList(protoexists: boolean; procname: string): integer;
    procedure ProcedureBlock;
//    procedure Error(s: string);
    procedure AbortMsg(s: string);
    procedure Expected(s: string);
    procedure Undefined(n: string);
    procedure CheckIdent;
    procedure Duplicate(n: string);
    function SizeOfType(dt: char): integer;
    procedure AddEntry(N: string; T: char);
    procedure CheckDup(N: string);
    procedure CheckTable(N: string);
    procedure AddParam(N: string; T: char);
    function DataType(n: string): char;
    procedure LoadVar(Name: string);
    procedure Store(Name: string);
    procedure Allocate(Name, Val: string; dt: char);
    procedure Epilog(bIsSub: boolean);
    procedure Header;
    procedure Prolog(name: string; bIsSub: boolean);
    procedure ReadIt(Name: string);
    procedure Trailer;
    procedure WriteIt;
//    procedure EmitTypeDefs;
    procedure EmitLn(s: string);
    procedure EmitLnNoTab(s: string);
    procedure PostLabel(L: string);
    procedure LoadConst(n: string);
//    procedure Negate;
    procedure NotIt;
    procedure PopAdd;
    procedure PopAnd;
    procedure PopCmpEqual;
    procedure PopCmpGreater;
    procedure PopCmpGreaterOrEqual;
    procedure PopCmpLess;
    procedure PopCmpLessOrEqual;
    procedure PopCmpNEqual;
    procedure PopDiv;
    procedure PopLeftShift;
    procedure PopMul;
    procedure PopOr;
    procedure PopRightShift;
    procedure PopSub;
    procedure PopXor;
    procedure PushPrim;
    procedure SetZeroCC;
    procedure Branch(L: string);
    procedure BranchFalse(L: string);
    procedure ClearReg;
    procedure ArrayAssignment(const name : string);
    procedure MathAssignment(const name : string);
    procedure StoreAdd(const name: string);
    procedure StoreDiv(const name: string);
    procedure StoreMod(const name: string);
    procedure StoreAnd(const name: string);
    procedure StoreOr(const name: string);
    procedure StoreXor(const name: string);
    procedure StoreAbs(const name: string);
    procedure StoreSign(const name: string);
    procedure StoreShift(bRight : boolean; const name: string);
    procedure StoreMul(const name: string);
    procedure StoreSub(const name: string);
    procedure StoreInc(const name: string; const val: integer = 1);
    procedure StoreDec(const name: string; const val: integer = 1);
    procedure DoWait;
    procedure DoAPICommands(const lend, lstart : string);
    procedure DoCls;
    procedure DoBreakContinue(idx : integer; const lbl : string);
    procedure DoOnFwdRev;
    procedure DoOnFwdRevReg;
    procedure DoOnFwdRevSync;
    procedure DoRotateMotors(idx : integer);
    procedure DoSetSensorTypeMode(idx : integer);
    procedure DoClearSetResetSensor;
    procedure DoReadSensor;
    procedure DoTextNumOut(idx : integer);
    procedure DoPlayTone;
    procedure DoPlayFile;
    procedure DoAcquireRelease;
    procedure DoResetSleepTimer;
    procedure DoPrecedesFollows;
    procedure DoReturn;
    procedure DoStopMotors;
    procedure ReportProblem(const lineNo: integer; const fName, line,
      msg: string; err: boolean);
    procedure Scan;
    function IsWhite(c: char): boolean;
    function IsRelop(c: char): boolean;
    function IsOrop(c: char): boolean;
    function IsAlpha(c: char): boolean;
    function IsDigit(c: char): boolean;
    function IsHex(c: char): boolean;
    function IsAlNum(c: char): boolean;
    function IsAddop(c: char): boolean;
    function IsMulop(c: char): boolean;
    procedure GetString;
    procedure CheckNumeric;
    procedure LoadAPIFunctions;
  protected
    fBadProgram : boolean;
    fProgErrorCount : integer;
    procedure InternalParseStream;
    procedure Clear;
    procedure PreProcess;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Parse(const aFilename : string); overload;
    procedure Parse(aStream : TStream); overload;
    procedure Parse(aStrings : TStrings); overload;
    property NBCSource : TStrings read fNBCSrc;
    property CompilerMessages : TStrings read fMessages;
    property IncludeDirs : TStrings read fIncludeDirs;
    property CurrentFile : string read fCurFile write fCurFile;
    property OnCompilerMessage : TOnCompilerMessage read fOnCompMSg write fOnCompMsg;
  end;

implementation

uses
  SysUtils, Math, uPreprocess, uNQCLexer {$IFDEF FAST_MM}, FastStrings{$ENDIF};

resourcestring
  sException    = 'Exception';
  sVariableName = 'Variable Name';
  sTooManyArgs  = 'Too many arguments';
  sMaxParamCountExceeded = 'Max param count exceeded';

{--------------------------------------------------------------}
{ Constant Declarations }

const
  TAB = ^I;
  CR  = ^M;
  LF  = ^J;

var
  LCount : integer = 0;

const
  MAXSYMSIZE   = 16;
  MAXGLOBALS   = 1000;
  MAXLOCALS    = 64;
  MAXSTAGE     = 128000;
  MAXDEFS      = 5000;
  MAXLINESIZE  = 80;
  MAXPARAMS    = 16;
  MAXCONSTS    = 100;
  MAXCONSTSIZE = 80;
  MAXINCL	     = 9;
  MAXSTACK     = 100;

{--------------------------------------------------------------}
{ Type Declarations }

type
  SymTab = array[1..MAXGLOBALS] of string;
  TabPtr = ^SymTab;
  ParTypeTab = array[1..MAXPARAMS] of char;

{--------------------------------------------------------------}
{ Variable Declarations }

var
  Look: char = LF;              { Lookahead Character }
  PrevLook : char;
  Token: char;             { Encoded Token       }
  Value: string;           { Unencoded Token     }

var
  slevel : integer = 1;
  linenumber : array[1..MAXINCL] of integer;	// current source line number
//  sname : array[1..MAXINCL] of string;// current source name
  totallines : integer = 0;
  endofallsource : boolean = False;

var
  GS_Name : SymTab;
  GS_Type : array[1..MAXGLOBALS] of char;
  GS_Size : array[1..MAXGLOBALS] of integer;	// size (in 'data type' units)
  NumGlobals : integer = 0;

  GS_ParamCount : array[1..MAXGLOBALS] of integer;
  GS_ParamType : array[1..MAXGLOBALS] of ParTypeTab;

  PS_Name : array[1..MAXPARAMS] of string;
  PS_Type : array[1..MAXPARAMS] of char;
  NumParams : integer = 0;

  LS_Name : array[1..MAXLOCALS] of string;
  LS_Type : array[1..MAXLOCALS] of char;
  NumLocals : integer = 0;
  TotLocSize : integer;

  CS_Name : array[1..MAXCONSTS] of string;	// constant symbol table (name)
  CS_Type : array[1..MAXCONSTS] of char;	// data type
  CS_Value : array[1..MAXCONSTS] of string;
  CCount : integer = 0;		// # of autogen constants
  NumConsts : integer = 0;		// total # of constants

{--------------------------------------------------------------}
{ Definition of Keywords and Token Types }

const TOK_IDENTIFIER		= 'x';
const TOK_IF			      = 'i';
const TOK_ELSE		      = 'l';
const TOK_DO            = 'd';
const TOK_ASM           = 'a';
const TOK_REPEAT        = 'r';
const TOK_SWITCH        = 's';
//const TOK_ENDIF		      = 'e';
const TOK_WHILE		      = 'w';
//const TOK_ENDWHILE		  = 'e';
const TOK_FOR			      = 'f';
//const TOK_ENDFOR		    = 'e';
const TOK_END			      = '}';
//const TOK_DEFINE		    = 'd';
//const TOK_INLINE		    = 'a';
//const TOK_PROGRAM		    = 'P';
const TOK_APIFUNC       = 'F';
const TOK_PROCEDURE		  = 'R';
const TOK_TASK          = 'K';
const TOK_BEGIN		      = '{';
//const TOK_INCLUDE		    = 'E';
const TOK_DIRECTIVE		  = '#';
const TOK_API           = 'Z';

const TOK_STRING        = 'S';
const TOK_NUM			      = 'N';
const TOK_HEX			      = 'H';
const TOK_UNSIGNED		  = 'U';
const TOK_CHARDEF		    = 'C';
const TOK_SHORTDEF 	    = 'I';
const TOK_LONGDEF       = 'L';
const TOK_BYTEDEF       = 'b';
const TOK_MUTEXDEF      = 'M';
const TOK_ULONGDEF      = #05;
const TOK_USHORTDEF     = #06;

const TOK_BLOCK_COMMENT = #01;
const TOK_LINE_COMMENT  = #02;
//const TOK_IFDEF	        = #03;
//const TOK_ENDIFDEF	    = #04;

const
  NKW  = 19; //18;
  NKW1 = 20; //19;

const
  KWlist: array[1..NKW] of string =
              ('if', 'else', {'endif', }'while', {'endwhile',}
               'for', {'endfor', }'sub', 'task',
               'do', 'repeat', 'switch', 'asm',
               'unsigned', 'long', 'short', 'int',
               'char', 'bool', 'byte', 'mutex',
               'end');

const                                     // 'xileweRWve'
  KWcode: array[1..NKW1+1] of Char =
    (TOK_IDENTIFIER, TOK_IF, TOK_ELSE,
		 {TOK_ENDIF, }TOK_WHILE, {TOK_ENDWHILE,}
     TOK_FOR, {TOK_ENDFOR, }TOK_PROCEDURE, TOK_TASK,
     TOK_DO, TOK_REPEAT, TOK_SWITCH, TOK_ASM,
     TOK_UNSIGNED, TOK_LONGDEF, TOK_SHORTDEF, TOK_SHORTDEF,
     TOK_CHARDEF, TOK_BYTEDEF, TOK_BYTEDEF, TOK_MUTEXDEF,
     TOK_END,
     #0);

const
  API_BREAK    = 0;
  API_CONTINUE = 1;
  API_READ     = 2;
  API_WRITE    = 3;
  API_WAIT     = 4;
  API_CLS      = 5;
  API_TEXTOUT  = 6;
  API_NUMOUT   = 7;
  API_ONFWD    = 8;
  API_ONREV    = 9;
  API_ONFWDREG = 10;
  API_ONREVREG = 11;
  API_ONFWDSYNC = 12;
  API_ONREVSYNC = 13;
  API_COAST     = 14;
  API_OFF       = 15;
  API_ROTATEMOTOR = 16;
  API_ROTATEMOTOREX = 17;
  API_SETSENSORTYPE = 18;
  API_SETSENSORMODE = 19;
  API_READSENSOR = 20;
  API_CLEARSENSOR = 21;
  API_SETSENSORTOUCH = 22;
  API_SETSENSORLIGHT = 23;
  API_SETSENSORSOUND = 24;
  API_SETSENSORULTRASOUND = 25;
  API_RESETSENSOR = 26;
  API_READSENSORUS = 27;
  API_PLAYTONE = 28;
  API_PLAYFILE = 29;
  API_RETURN = 30;
  API_ACQUIRE = 31;
  API_RELEASE = 32;
  API_PRECEDES = 33;
  API_FOLLOWS  = 34;
  API_RESETSLEEP = 35;

  APICount = 36;
  APIList : array[0..APICount-1] of string = (
    'break', 'continue',
    'Read', 'Write', 'Wait', 'Cls', 'TextOut', 'NumOut',
    'OnFwd', 'OnRev', 'OnFwdReg', 'OnRevReg', 'OnFwdSync',
    'OnRevSync', 'Coast', 'Off', 'RotateMotor', 'RotateMotorEx',
    'SetSensorType', 'SetSensorMode', 'ReadSensor', 'ClearSensor',
    'SetSensorTouch', 'SetSensorLight', 'SetSensorSound', 'SetSensorUltrasonic',
    'ResetSensor', 'ReadSensorUS', 'PlayTone', 'PlayFile',
    'return', 'Acquire', 'Release', 'Precedes', 'Follows', 'ResetSleepTimer'
  );

var
//  outp : TextFile;
  sd : integer;
  maxsd : integer;

function tos : string;
begin
  Result := Format('__stack_%3.3d', [sd]);
end;

procedure pop;
begin
  dec(sd);
end;

procedure push;
begin
  inc(sd);
  maxsd := Max(maxsd, sd);
end;


procedure TTinyComp.GetCharX;
var
  bytesread : integer;
begin
  PrevLook := Look;
  bytesread := fMS.Read(Look, 1);
//  Read(inp, Look);
  fCurrentLine := fCurrentLine + Look;
  if Look = LF then
  begin
    IncLineNumber;
    fCurrentLine := '';
  end;
  if bytesread < 1 then
    endofallsource := True;
  if endofallsource and (slevel > 1) then begin
    // close file pointer
    linenumber[slevel] := 0;
    dec(slevel);
    Look := LF;
    endofallsource := False;
  end;
end;

{--------------------------------------------------------------}
{ Read New Character From Input Stream }

procedure TTinyComp.GetChar;
begin
  if fTempChar <> ' ' then begin
    Look := fTempChar;
    fTempChar := ' ';
  end
  else begin
    GetCharX;
    if Look = '/' then begin
      fMS.Read(fTempChar, 1);
//      Read(inp, tempchar);
      if fTempChar = '*' then begin
        Look := TOK_BLOCK_COMMENT;
        fTempChar := ' ';
      end
      else if fTempChar = '/' then begin
        Look := TOK_LINE_COMMENT;
        fTempChar := ' ';
      end;
    end;
  end;
end;


(*
{--------------------------------------------------------------}
{ Report an Error }

procedure TTinyComp.Error(s: string);
begin
  fMessages.Add(Format('Error: %s.', [s]));
end;
*)

{--------------------------------------------------------------}
{ Report Error and Halt }

procedure TTinyComp.ReportProblem(const lineNo: integer; const fName, line,
  msg: string; err : boolean);
var
  tmp, tmp1, tmp2, tmp3, tmp4 : string;
  stop : boolean;
begin
  if lineNo = -1 then
  begin
    tmp := msg;
    fMessages.Add(tmp);
  end
  else
  begin
    if err then
      tmp1 := Format('# Error: %s', [msg])
    else
      tmp1 := Format('# Warning: %s', [msg]);
    fMessages.Add(tmp1);
    tmp2 := Format('File "%s" ; line %d', [fName, lineNo]);
    fMessages.Add(tmp2);
    tmp3 := Format('#   %s', [fCurrentLine]);
    fMessages.Add(tmp3);
    tmp4 := '#----------------------------------------------------------';
    fMessages.Add(tmp4);
    tmp := tmp1+#13#10+tmp2+#13#10+tmp3+#13#10+tmp4;
  end;
  fBadProgram := err;
  if err then
    inc(fProgErrorCount);
  stop := false;
  if assigned(fOnCompMsg) then
    fOnCompMsg(tmp, stop);
  if stop then
    Abort;
end;

procedure TTinyComp.AbortMsg(s: string);
begin
  ReportProblem(linenumber[slevel], CurrentFile, Value, s, True);
//  Next; // advance parser
//  Abort;
end;

{--------------------------------------------------------------}
{Return the size in base units of a standard datatype }

function TTinyComp.SizeOfType(dt : char) : integer;
begin
  case dt of
    TOK_CHARDEF, TOK_BYTEDEF : Result := 1;
    TOK_SHORTDEF, TOK_USHORTDEF : Result := 2;
    TOK_LONGDEF, TOK_ULONGDEF : Result := 4;
    TOK_MUTEXDEF : Result := 4;
  else
    Result := 0;
    AbortMsg('SizeOfType() - Unknown Data Type');
  end;
end;

{--------------------------------------------------------------}
{ Report What Was Expected }

procedure TTinyComp.Expected(s: string);
begin
  AbortMsg(s + ' Expected');
end;

{--------------------------------------------------------------}
{ Report an Undefined Identifier }

procedure TTinyComp.Undefined(n: string);
begin
  AbortMsg('Undefined Identifier ' + n);
end;


{--------------------------------------------------------------}
{ Report a Duplicate Identifier }

procedure TTinyComp.Duplicate(n: string);
begin
  AbortMsg('Duplicate Identifier ' + n);
end;


{--------------------------------------------------------------}
{ Check to Make Sure the Current Token is an Identifier }

procedure TTinyComp.CheckIdent;
begin
  if Token <> TOK_IDENTIFIER then Expected('Identifier');
end;

{--------------------------------------------------------------}
{ Check to Make Sure the Current Token is a Number }

procedure TTinyComp.CheckNumeric;
begin
  if not (Token in [TOK_NUM, TOK_HEX]) then Expected('Number');
end;



{--------------------------------------------------------------}
{ Recognize an Alpha Character }

function TTinyComp.IsAlpha(c: char): boolean;
begin
  Result := upcase(c) in ['A'..'Z'];
end;

{--------------------------------------------------------------}
{ Recognize a Decimal Digit }

function TTinyComp.IsDigit(c: char): boolean;
begin
  Result := c in ['0'..'9'];
end;

{--------------------------------------------------------------}
{ Recognize a Hex Digit }

function TTinyComp.IsHex(c: char): boolean;
begin
  Result := IsDigit(c) or (c in ['a'..'f', 'A'..'F']);
end;

{--------------------------------------------------------------}
{ Recognize an Alphanumeric }

function TTinyComp.IsAlNum(c: char): boolean;
begin
  Result := IsAlpha(c) or IsDigit(c) or (c = '_');
end;

{--------------------------------------------------------------}
{ Recognize an Addop }

function TTinyComp.IsAddop(c: char) : boolean;
begin
  Result := c in ['+', '-'];
end;

{--------------------------------------------------------------}
{ Recognize a Mulop }

function TTinyComp.IsMulop(c: char): boolean;
begin
  Result := c in ['*', '/'];
end;

{--------------------------------------------------------------}
{ Recognize a Boolean Orop }

function TTinyComp.IsOrop(c: char): boolean;
begin
  Result := c in ['|', '~'];
end;

{--------------------------------------------------------------}
{ Recognize a Relop }

function TTinyComp.IsRelop(c: char): boolean;
begin
  Result := c in ['=', '!', '<', '>'];
end;

{--------------------------------------------------------------}
{ Recognize White Space }

function TTinyComp.IsWhite(c: char): boolean;
begin
//  if c = LF then
//    IncLineNumber;
  Result := c in [' ', TAB, CR, LF, TOK_BLOCK_COMMENT, TOK_LINE_COMMENT];
end;

{--------------------------------------------------------------}
{ Skip A Comment Field }

procedure TTinyComp.SkipCommentBlock;
begin
  repeat
    repeat
      GetCharX;
//      if Look = LF then
//        IncLineNumber;
    until (Look = '*') or endofallsource;
    GetCharX;
//    if Look = LF then
//      IncLineNumber;
  until (Look = '/') or endofallsource;
  GetChar;
end;


{--------------------------------------------------------------}
{ Skip A Comment To End Of Line field }

procedure TTinyComp.SkipCommentLine;
begin
  repeat
    GetCharX;
  until (Look = LF) or endofallsource;
//  if Look = LF then
//    IncLineNumber;
  GetChar;
end;

procedure TTinyComp.SkipDirectiveLine;
begin
  fDirLine := '#' + Value + ' ';
  repeat
    GetCharX;
    fDirLine := fDirLine + Look;
  until (Look = LF) or endofallsource;
//  if Look = LF then
//    IncLineNumber;
  GetChar;
//  Next;
end;

{--------------------------------------------------------------}
{ Skip Over Leading White Space }

procedure TTinyComp.SkipWhite;
begin
  while IsWhite(Look) and not endofallsource do begin
    case Look of
      TOK_LINE_COMMENT : SkipCommentLine;
      TOK_BLOCK_COMMENT : SkipCommentBlock;
    else
      GetChar;
    end;
  end;
end;


{--------------------------------------------------------------}
{ Table Lookup }

function Lookup(T: TabPtr; s: string; n: integer): integer;
var
  i: integer;
  found: Boolean;
begin
  found := false;
  i := n;
  while (i > 0) and not found do
     if s = T^[i] then
        found := true
     else
        dec(i);
  Result := i;
end;


{--------------------------------------------------------------}
{ Locate a Symbol in Table }
{ Returns the index of the entry.  Zero if not present. }

function Locate(n: string): integer;
begin
  Result := Lookup(@GS_Name, n, NumGlobals);
end;


{--------------------------------------------------------------}
{ Look for Symbol in Table }

function InTable(n: string): Boolean;
begin
  Result := Lookup(@GS_Name, n, NumGlobals) <> 0;
end;


{--------------------------------------------------------------}
{ Look for Symbol in Parameter Table }

function IsParam(n: string): integer;
begin
  Result := Lookup(@PS_Name, n, NumParams);
end;


{--------------------------------------------------------------}
{ Look for Symbol in Local Table }

function IsLocal(n: string): integer;
begin
  Result := Lookup(@LS_Name, n, NumLocals);
end;


{--------------------------------------------------------------}
{ Look for Symbol in Constants Table }

function IsConst(n: string): integer;
begin
  Result := Lookup(@CS_Name, n, NumConsts);
end;


{--------------------------------------------------------------}
{ Test to see if name is already known }

function IsKnown(n : string) : integer;
begin
  Result := Locate(n) or IsLocal(n) or IsParam(n) or IsConst(n);
end;

{--------------------------------------------------------------}
{ Check to See if an Identifier is in the Symbol Table         }
{ Report an error if it's not. }

procedure TTinyComp.CheckTable(N: string);
begin
  if not InTable(N) then Undefined(N);
end;


{--------------------------------------------------------------}
{ Check the Symbol Table for a Duplicate Identifier }
{ Report an error if identifier is already in table. }


procedure TTinyComp.CheckDup(N: string);
begin
  if InTable(N) then Duplicate(N);
end;


{--------------------------------------------------------------}
{ Add a New Entry to Symbol Table }

procedure TTinyComp.AddEntry(N: string; T: char);
begin
  CheckDup(N);
  if NumGlobals = MAXGLOBALS then AbortMsg('Symbol Table Full');
  Inc(NumGlobals);
  GS_Name[NumGlobals] := N;
  GS_Type[NumGlobals] := T;
end;


{--------------------------------------------------------------}
{ Get an Identifier }

procedure TTinyComp.GetName;
begin
  SkipWhite;
  if Not IsAlpha(Look) then Expected('Identifier');
  Token := TOK_IDENTIFIER;
  Value := '';
  repeat
    Value := Value + Look;
//    Value := Value + UpCase(Look);
    GetChar;
  until not IsAlNum(Look);
end;


{--------------------------------------------------------------}
{ Get a Number }

procedure TTinyComp.GetNum;
var
  savedLook : char;
begin
  SkipWhite;
  if not IsDigit(Look) then Expected('Number');
  savedLook := Look;
  GetChar;
  if Look in ['x', 'X'] then
  begin
    GetHexNum;
  end
  else
  begin
    Token := TOK_NUM;
    Value := savedLook;
    if not IsDigit(Look) then Exit;
    repeat
      Value := Value + Look;
      GetChar;
    until not IsDigit(Look);
  end;
end;


{--------------------------------------------------------------}
{ Get a Hex Number }

procedure TTinyComp.GetHexNum;
begin
  SkipWhite;
  GetChar(); // skip the $ (or 'x')
  if not IsHex(Look) then Expected('Hex number');
  Token := TOK_HEX;
  Value := '0x';
  repeat
    Value := Value + Look;
    GetChar;
  until not IsHex(Look);
end;


{--------------------------------------------------------------}
{ Get a Character Literal }

procedure TTinyComp.GetCharLit;
begin
  SkipWhite;
  GetChar; // skip the '
  Token := TOK_NUM;
  Value := IntToStr(Ord(Look));
  GetChar;
  if Look <> '''' then Expected('Character literal');
  GetChar;
end;

{--------------------------------------------------------------}
{ Get a string Literal }

procedure TTinyComp.GetString;
begin
  SkipWhite;
  GetChar; // skip the "
  Token := TOK_STRING;
  Value := '''' + Look;
  repeat
    GetCharX;
    if (Look <> LF) and (Look <> '"') then
      Value := Value + Look;
  until (Look = '"') or (Look = LF) or endofallsource;
  Value := Value + '''';
  if Look <> '"' then Expected('String literal');
  GetChar; 
end;


{--------------------------------------------------------------}
{ Get an Operator }

procedure TTinyComp.GetOp;
begin
  SkipWhite;
  Token := Look;
  Value := Look;
  GetChar;
end;


{--------------------------------------------------------------}
{ Get the Next Input Token }

procedure TTinyComp.Next;
begin
  SkipWhite;
  if Look = '''' then GetCharLit
  else if Look = '"' then GetString
  else if IsAlpha(Look) then GetName
  else if IsDigit(Look) then GetNum
  else if Look = '$' then GetHexNum
  else GetOp;
end;

function IsAPICommand(const name : string) : boolean;
var
  i : integer;
begin
  Result := False;
  for i := Low(APIList) to High(APIList) do
  begin
    if APIList[i] = name then
    begin
      Result := True;
      Break;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Scan the Current Identifier for Keywords }

procedure TTinyComp.Scan;
var
  idx : integer;
begin
  if Token = TOK_IDENTIFIER then
  begin
    idx := Lookup(Addr(KWlist), Value, NKW);
    if idx <> 0 then
      Token := KWcode[idx + 1]
    else
    begin
      // is it an API command?
      if IsAPICommand(Value) then
        Token := TOK_API;
    end;
  end;
end;


{--------------------------------------------------------------}
{ Match a Specific Input String }

procedure TTinyComp.MatchString(x: string);
begin
  if Value <> x then Expected('''' + x + '''');
  Next;
end;


{--------------------------------------------------------------}
{ Match a Semicolon }

procedure TTinyComp.Semi;
begin
//  MatchString(';');
  if Token = ';' then
    Next;
end;


{--------------------------------------------------------------}
{ Output a String with Tab and CRLF }

procedure TTinyComp.EmitLn(s: string);
begin
  fNBCSrc.Add(#9+s);
end;


{--------------------------------------------------------------}
{ Output a String without Tab with CRLF }

procedure TTinyComp.EmitLnNoTab(s: string);
begin
  fNBCSrc.Add(s);
end;


{--------------------------------------------------------------}
{ Generate a Unique Label }

function NewLabel: string;
var
  S: string;
begin
  Str(LCount, S);
  NewLabel := 'L' + S;
  Inc(LCount);
end;


{--------------------------------------------------------------}
{ Post a Label To Output }

procedure TTinyComp.PostLabel(L: string);
begin
  EmitLnNoTab(L+':');
end;


{---------------------------------------------------------------}
{ Initialize Parameter Table to Null }
procedure ClearParams;
var
  i : integer;
begin
  for i := 1 to MAXPARAMS do
    PS_Name[i] := '';
  NumParams := 0;
end;

{---------------------------------------------------------------}
{ Initialize Locals Table to Null }
procedure ClearLocals;
var
  i : integer;
begin
  for i := 1 to MAXLOCALS do begin
    LS_Name[i] := '';
    LS_Type[i] := #0;
  end;
  NumLocals := 0;
  TotLocSize := 0;
end;

{--------------------------------------------------------------}
{ Add a Parameter to Table }

procedure TTinyComp.AddParam(N: string; T: char);
begin
  if IsParam(N) <> 0 then Duplicate(N);
  Inc(NumParams);
  PS_Name[NumParams] := N;
  PS_Type[NumParams] := T;
end;

type
  TSymbolType = (stUnknown, stParam, stLocal, stGlobal);

function WhatIs(n : string) : TSymbolType;
begin
  if IsParam(n) <> 0 then Result := stParam
  else if IsLocal(n) <> 0 then Result := stLocal
  else if InTable(n) then Result := stGlobal
  else Result := stUnknown;
end;

function TTinyComp.DataType(n : string) : char;
begin
  case WhatIs(n) of
    stParam : Result := PS_Type[IsParam(n)];
    stLocal : Result := LS_Type[IsLocal(n)];
    stGlobal : Result := GS_Type[Locate(n)];
  else
    Result := #0;
    AbortMsg('Unknown data type');
  end;
end;

{---------------------------------------------------------------}
{ Add Primary or var }

procedure TTinyComp.StoreAdd(const name : string);
begin
  EmitLn(Format('add %0:s, %0:s, __D0', [name]));
end;

{---------------------------------------------------------------}
{ Subtract Primary from var }

procedure TTinyComp.StoreSub(const name : string);
begin
  EmitLn(Format('sub %0:s, %0:s, __D0', [name]));
end;

{---------------------------------------------------------------}
{ Multiply Primary with var }

procedure TTinyComp.StoreMul(const name : string);
begin
  EmitLn(Format('mul %0:s, %0:s, __D0', [name]));
end;

{---------------------------------------------------------------}
{ Divide Primary with var }

procedure TTinyComp.StoreDiv(const name : string);
begin
  EmitLn(Format('div %0:s, %0:s, __D0', [name]));
end;

{---------------------------------------------------------------}
{ Mod Primary with var }

procedure TTinyComp.StoreMod(const name : string);
begin
  EmitLn(Format('mod %0:s, %0:s, __D0', [name]));
end;

procedure TTinyComp.StoreAbs(const name: string);
begin
  EmitLn(Format('abs %0:s, __D0', [name]));
end;

procedure TTinyComp.StoreAnd(const name: string);
begin
  EmitLn(Format('and %0:s, %0:s, __D0', [name]));
end;

procedure TTinyComp.StoreOr(const name: string);
begin
  EmitLn(Format('or %0:s, %0:s, __D0', [name]));
end;

procedure TTinyComp.StoreShift(bRight: boolean; const name: string);
begin
// generate code for left and right shift
end;

procedure TTinyComp.StoreSign(const name: string);
begin
  EmitLn(Format('sign %0:s, __D0', [name]));
end;

procedure TTinyComp.StoreXor(const name: string);
begin
  EmitLn(Format('xor %0:s, %0:s, __D0', [name]));
end;

{---------------------------------------------------------------}
{ increment var }

procedure TTinyComp.StoreInc(const name : string; const val : integer);
begin
  EmitLn(Format('add %0:s, %0:s, %d', [name, val]));
end;

{---------------------------------------------------------------}
{ decrement var }

procedure TTinyComp.StoreDec(const name : string; const val : integer);
begin
  EmitLn(Format('sub %0:s, %0:s, %d', [name, val]));
end;

{---------------------------------------------------------------}
{ Clear the Primary Register }

procedure TTinyComp.ClearReg;
begin
  EmitLn('set __D0,0');
end;

(*
{---------------------------------------------------------------}
{ Negate the Primary Register }

procedure TTinyComp.Negate;
begin
  EmitLn('neg __D0,__D0');
end;
*)

{---------------------------------------------------------------}
{ Complement the Primary Register }

procedure TTinyComp.NotIt;
begin
  EmitLn('not __D0, __D0');
end;


{---------------------------------------------------------------}
{ Load a Constant Value to Primary Register }

procedure TTinyComp.LoadConst(n: string);
var
  cval : int64;
  tmpSrc : string;
begin
  cval := StrToIntDef(n, 0);
  if (cval < Low(smallint)) or (cval > High(word)) then
    tmpSrc := 'mov __D0, %s'
  else
    tmpSrc := 'set __D0, %s';
  EmitLn(Format(tmpSrc, [n]));
end;

{---------------------------------------------------------------}
{ Load a Variable to Primary Register }

procedure TTinyComp.LoadVar(Name: string);
var
  dt : char;
begin
  dt := DataType(Name);
  if dt = TOK_PROCEDURE then AbortMsg('Can not assign to/from a procedure');
  case WhatIs(Name) of
    stParam : begin
    end;
    stLocal : begin
    end;
    stGlobal : begin
      EmitLn('mov __D0,' + Name);
    end;
    stUnknown : Undefined(Name);
  else
    AbortMsg('Unsupported LoadVar mechanism');
  end;
end;

{---------------------------------------------------------------}
{ Push Primary onto Stack }

procedure TTinyComp.PushPrim;
begin
  push;
  EmitLn('mov ' + tos + ',__D0');
end;

{---------------------------------------------------------------}
{ Add Top of Stack to Primary }

procedure TTinyComp.PopAdd;
begin
  EmitLn('add __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Subtract Primary from Top of Stack }

procedure TTinyComp.PopSub;
begin
  EmitLn('sub __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Multiply Top of Stack by Primary }

procedure TTinyComp.PopMul;
begin
  EmitLn('mul __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Divide Top of Stack by Primary }

procedure TTinyComp.PopDiv;
begin
  EmitLn('div __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ AND Top of Stack with Primary }

procedure TTinyComp.PopAnd;
begin
  EmitLn('and __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ OR Top of Stack with Primary }

procedure TTinyComp.PopOr;
begin
  EmitLn('or __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ XOR Top of Stack with Primary }

procedure TTinyComp.PopXor;
begin
  EmitLn('xor __D0,' + tos + ',__D0');
  pop;
end;


{---------------------------------------------------------------}
{ Left Shift Top of Stack to Primary }

procedure TTinyComp.PopLeftShift;
begin
  EmitLn('shl __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Right Shift Top of Stack to Primary }

procedure TTinyComp.PopRightShift;
begin
  EmitLn('shr __D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set zero flag based on __D0 value }

procedure TTinyComp.SetZeroCC;
begin
  fCCSet := True;
  EmitLn('tst NEQ,__zf,__D0');
end;


{---------------------------------------------------------------}
{ Set __D0 If Compare was = }

procedure TTinyComp.PopCmpEqual;
begin
  fCCSet := True;
  EmitLn('cmp EQ,__zf,' + tos + ',__D0');
//  EmitLn('cmp EQ,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was != }

procedure TTinyComp.PopCmpNEqual;
begin
  fCCSet := True;
  EmitLn('cmp NEQ,__zf,' + tos + ',__D0');
//  EmitLn('cmp NEQ,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was > }

procedure TTinyComp.PopCmpGreater;
begin
  fCCSet := True;
  EmitLn('cmp GT,__zf,' + tos + ',__D0');
//  EmitLn('cmp GT,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was < }

procedure TTinyComp.PopCmpLess;
begin
  fCCSet := True;
  EmitLn('cmp LT,__zf,' + tos + ',__D0');
//  EmitLn('cmp LT,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was <= }

procedure TTinyComp.PopCmpLessOrEqual;
begin
  fCCSet := True;
  EmitLn('cmp LTEQ,__zf,' + tos + ',__D0');
//  EmitLn('cmp LTEQ,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Set __D0 If Compare was >= }

procedure TTinyComp.PopCmpGreaterOrEqual;
begin
  fCCSet := True;
  EmitLn('cmp GTEQ,__zf,' + tos + ',__D0');
//  EmitLn('cmp GTEQ,__D0,' + tos + ',__D0');
  pop;
end;

{---------------------------------------------------------------}
{ Store Primary to Variable }

procedure TTinyComp.Store(Name: string);
begin
  if DataType(Name) = TOK_PROCEDURE then
    AbortMsg('Can not assign to/from a procedure');
  case WhatIs(Name) of
    stParam : begin
    end;
    stLocal : begin
    end;
    stGlobal : begin
      EmitLn('mov ' + Name + ',__D0');
    end;
    stUnknown : Undefined(Name);
  else
    AbortMsg('Unsupported Store mechanism');
  end;
end;

{---------------------------------------------------------------}
{ Branch Unconditional  }

procedure TTinyComp.Branch(L: string);
begin
  EmitLn('jmp ' + L);
end;

{---------------------------------------------------------------}
{ Branch False }

procedure TTinyComp.BranchFalse(L: string);
begin
  // if the condition code has not been set then set it manually
  if not fCCSet then
    SetZeroCC;
  EmitLn('brtst EQ,' + L + ',__zf');
//  EmitLn('brtst EQ,' + L + ',__D0');
end;

{---------------------------------------------------------------}
{ Read Variable to Primary Register }

procedure TTinyComp.ReadIt(Name: string);
begin
  EmitLn(';BSR READ');
  Store(Name);
end;


{ Write from Primary Register }

procedure TTinyComp.WriteIt;
begin
  EmitLn('set __writeDTArgs.Options, 1');
  EmitLn('set __writeDTArgs.Location.X, 0');
  EmitLn('set __writeDTArgs.Location.Y, 56');
  EmitLn('numtostr __writeDTArgs.Text, __D0');
  EmitLn('syscall DrawText, __writeDTArgs');
end;


{--------------------------------------------------------------}
{ Write Header Info }

procedure TTinyComp.Header;
begin
  // do nothing
end;

(*
const
  STANDARD_TYPEDEFS =
    'dseg segment'#13#10 +
    '  TLocation	struct'#13#10 +
    '   X		sword'#13#10 +
    '   Y		sword'#13#10 +
    '  TLocation	ends'#13#10 +
    '  TSize	struct'#13#10 +
    '   Width	sword'#13#10 +
    '   Height	sword'#13#10 +
    '  TSize	ends'#13#10 +
    '  TFileOpen	struct'#13#10 +
    '   Result		word'#13#10 +
    '   FileHandle	byte'#13#10 +
    '   Filename	byte[]'#13#10 +
    '   Length		dword'#13#10 +
    '  TFileOpen	ends'#13#10 +
    '  TFileReadWrite	struct'#13#10 +
    '   Result		word'#13#10 +
    '   FileHandle	byte'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '   Length		dword'#13#10 +
    '  TFileReadWrite	ends'#13#10 +
    '  TFileClose	struct'#13#10 +
    '   Result		word'#13#10 +
    '   FileHandle	byte'#13#10 +
    '  TFileClose	ends'#13#10 +
    '  TFileResolveHandle	struct'#13#10 +
    '   Result		word'#13#10 +
    '   FileHandle	byte'#13#10 +
    '   WriteHandle	byte'#13#10 +
    '   Filename	byte[]'#13#10 +
    '  TFileResolveHandle	ends'#13#10 +
    '  TFileRename	struct'#13#10 +
    '   Result		word'#13#10 +
    '   OldFilename	byte[]'#13#10 +
    '   NewFilename	byte[]'#13#10 +
    '  TFileRename	ends'#13#10 +
    '  TFileDelete	struct'#13#10 +
    '   Result		word'#13#10 +
    '   Filename	byte[]'#13#10 +
    '  TFileDelete	ends'#13#10 +
    '  TSoundPlayFile	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Filename	byte[]'#13#10 +
    '   Loop		byte'#13#10 +
    '   Volume		byte'#13#10 +
    '  TSoundPlayFile	ends'#13#10 +
    '  TSoundPlayTone	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Frequency	word'#13#10 +
    '   Duration	word'#13#10 +
    '   Loop		byte'#13#10 +
    '   Volume		byte'#13#10 +
    '  TSoundPlayTone	ends'#13#10 +
    '  TSoundGetState	struct'#13#10 +
    '   State		byte'#13#10 +
    '   Flags		byte'#13#10 +
    '  TSoundGetState	ends'#13#10 +
    '  TSoundSetState	struct'#13#10 +
    '   Result		byte'#13#10 +
    '   State		byte'#13#10 +
    '   Flags		byte'#13#10 +
    '  TSoundSetState	ends'#13#10 +
    '  TDrawText	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Location	TLocation'#13#10 +
    '   Text		byte[]'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawText	ends'#13#10 +
    '  TDrawPoint	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Location	TLocation'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawPoint	ends'#13#10 +
    '  TDrawLine	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   StartLoc	TLocation'#13#10 +
    '   EndLoc		TLocation'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawLine	ends'#13#10 +
    '  TDrawCircle	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Center		TLocation'#13#10 +
    '   Size		byte'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawCircle	ends'#13#10 +
    '  TDrawRect	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Location	TLocation'#13#10 +
    '   Size		TSize'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawRect	ends'#13#10 +
    '  TDrawGraphic	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Location	TLocation'#13#10 +
    '   Filename	byte[]'#13#10 +
    '   Variables	sdword[]'#13#10 +
    '   Options	dword'#13#10 +
    '  TDrawGraphic	ends'#13#10 +
    '  TSetScreenMode	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   ScreenMode	dword'#13#10 +
    '  TSetScreenMode	ends'#13#10 +
    '  TReadButton	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Index		byte'#13#10 +
    '   Pressed	byte'#13#10 +
    '   Count		byte'#13#10 +
    '   Reset		byte'#13#10 +
    '  TReadButton	ends'#13#10 +
    '  TCommLSWrite	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Port		byte'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '   ReturnLen	byte'#13#10 +
    '  TCommLSWrite	ends'#13#10 +
    '  TCommLSRead	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Port		byte'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '   BufferLen	byte'#13#10 +
    '  TCommLSRead	ends'#13#10 +
    '  TCommLSCheckStatus	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Port		byte'#13#10 +
    '   BytesReady	byte'#13#10 +
    '  TCommLSCheckStatus	ends'#13#10 +
    '  TRandomNumber	struct'#13#10 +
    '   Result		sword'#13#10 +
    '  TRandomNumber	ends'#13#10 +
    '  TGetStartTick	struct'#13#10 +
    '   Result		dword'#13#10 +
    '  TGetStartTick	ends'#13#10 +
    '  TMessageWrite	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   QueueID	byte'#13#10 +
    '   Message	byte[]'#13#10 +
    '  TMessageWrite	ends'#13#10 +
    '  TMessageRead	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   QueueID	byte'#13#10 +
    '   Remove		byte'#13#10 +
    '   Message	byte[]'#13#10 +
    '  TMessageRead	ends'#13#10 +
    '  TCommBTWrite	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Connection	byte'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '  TCommBTWrite	ends'#13#10 +
    '  TCommBTRead	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Count		byte'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '  TCommBTRead	ends'#13#10 +
    '  TCommBTCheckStatus	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   Connection	byte'#13#10 +
    '  TCommBTCheckStatus	ends'#13#10 +
    '  TKeepAlive	struct'#13#10 +
    '   Result		dword'#13#10 +
    '  TKeepAlive	ends'#13#10 +
    '  TIOMapRead	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   ModuleName	byte[]'#13#10 +
    '   Offset		word'#13#10 +
    '   Count		word'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '  TIOMapRead	ends'#13#10 +
    '  TIOMapWrite	struct'#13#10 +
    '   Result		sbyte'#13#10 +
    '   ModuleName	byte[]'#13#10 +
    '   Offset		word'#13#10 +
    '   Buffer		byte[]'#13#10 +
    '  TIOMapWrite	ends'#13#10 +
    'dseg ends';

procedure TTinyComp.EmitTypeDefs;
begin
  fNBCSrc.Text := STANDARD_TYPEDEFS;
end;
*)

{--------------------------------------------------------------}
{ Write Trailer Info }

procedure TTinyComp.Trailer;
var
  i : integer;
  tmp : TStrings;
begin
  // handle stack variables
  tmp := TStringList.Create;
  try
    tmp.AddStrings(fNBCSrc);
    fNBCSrc.Clear;
//    EmitTypeDefs;
    EmitLnNoTab('#include "NXTDefs.h"');
    EmitLnNoTab('dseg segment');
    EmitLn('__D0 slong');
    EmitLn('__zf byte');
    EmitLn('__writeDTArgs TDrawText');
    EmitLn('__clsSSMArgs TSetScreenMode');
    for i := 1 to maxsd do
      EmitLn(Format('__stack_%3.3d slong', [i]));
    EmitLnNoTab('dseg ends');
    fNBCSrc.AddStrings(tmp);
  finally
    tmp.Free;
  end;
end;

{--------------------------------------------------------------}
{ Write the Prolog }

procedure TTinyComp.Prolog(name : string; bIsSub : boolean);
begin
  if bIsSub then
    EmitLnNoTab('subroutine ' + name)
  else
    EmitLnNoTab('thread ' + name);
end;

{--------------------------------------------------------------}
{ Write the Epilog }

procedure TTinyComp.Epilog(bIsSub : boolean);
begin
  if bIsSub then
  begin
    EmitLn('return');
    EmitLnNoTab('ends');
  end
  else
    EmitLnNoTab('endt');
  EmitLnNoTab('');
end;

{--------------------------------------------------------------}
{ Allocate Storage for a Static variable }

procedure TTinyComp.Allocate(Name, Val: string; dt : char);
begin
  EmitLnNoTab('dseg segment');
  case dt of
    TOK_CHARDEF		: EmitLn(Format('%s sbyte %s', [Name, Val]));
    TOK_SHORTDEF 	: EmitLn(Format('%s sword %s', [Name, Val]));
    TOK_LONGDEF   : EmitLn(Format('%s sdword %s', [Name, Val]));
    TOK_BYTEDEF   : EmitLn(Format('%s byte %s', [Name, Val]));
    TOK_USHORTDEF : EmitLn(Format('%s word %s', [Name, Val]));
    TOK_ULONGDEF  : EmitLn(Format('%s dword %s', [Name, Val]));
    TOK_MUTEXDEF  : EmitLn(Format('%s mutex', [Name]));
  else
    EmitLn(Format('%s sword %s', [Name, Val]));
  end;
  EmitLnNoTab('dseg ends');
end;

{---------------------------------------------------------------}
{ Parse and Translate a Math Factor }

procedure TTinyComp.Factor;
var
  savedtoken : char;
  savedvalue : string;
begin
  if Token = '(' then begin
    Next;
    BoolExpression;
    MatchString(')');
  end
  else begin
    savedtoken := Token;
    savedvalue := Value;
    Next;
    case savedtoken of
      TOK_IDENTIFIER : begin
        if Token = '[' then
        begin
          Next;
          Expression;
          MatchString(']');
          EmitLn('index __D0, ' + savedvalue + ', __D0');
//          EmitLn('; copy specified element from array '+savedvalue);
//          EmitLn('; into __D0 (our favorite register)');
        end
        else if savedvalue = 'true' then
          LoadConst('-1')
        else if savedvalue = 'false' then
          LoadConst('0')
        else if fAPIFunctions.IndexOf(savedvalue) <> -1 then
        begin
          DoCallAPIFunc(savedvalue);
        end
        else
          LoadVar(savedvalue);
      end;
      TOK_NUM, TOK_HEX : begin
        LoadConst(savedvalue);
      end;
    else
      Expected('Math Factor');
    end;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate a Multiply }

procedure TTinyComp.Multiply;
begin
  Next;
  Factor;
  PopMul;
end;

{-------------------------------------------------------------}
{ Recognize and Translate a Divide }

procedure TTinyComp.Divide;
begin
  Next;
  Factor;
  PopDiv;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Math Term }

procedure TTinyComp.Term;
begin
  Factor;
  while IsMulop(Token) do begin
    PushPrim;
    case Token of
      '*': Multiply;
      '/': Divide;
    end;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate an Add }

procedure TTinyComp.Add;
begin
  Next;
  Term;
  PopAdd;
end;

{-------------------------------------------------------------}
{ Recognize and Translate a Subtract }

procedure TTinyComp.Subtract;
begin
  Next;
  Term;
  PopSub;
end;

{---------------------------------------------------------------}
{ Parse and Translate an Expression }

procedure TTinyComp.Expression;
begin
  if IsAddOp(Token) then
    ClearReg
  else
    Term;
  while IsAddop(Token) do begin
    PushPrim;
    case Token of
      '+': Add;
      '-': Subtract;
    end;
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Equals" }

procedure TTinyComp.Equal;
begin
  Next; // two equal signs of equality comparison
  Next;
  Expression;
  PopCmpEqual;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Less Than or Equal" }

procedure TTinyComp.LessOrEqual;
begin
  Next;
  Expression;
  PopCmpLessOrEqual;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Not Equals" }

procedure TTinyComp.NotEqual;
begin
  Next;
  Expression;
  PopCmpNEqual;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a left shift }

procedure TTinyComp.LeftShift;
begin
  Next;
  Expression;
  PopLeftShift;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a right shift }

procedure TTinyComp.RightShift;
begin
  Next;
  Expression;
  PopRightShift;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Less Than" }

procedure TTinyComp.Less;
begin
  Next;
  case Token of
    '=' : LessOrEqual;
    '>' : NotEqual;
    '<' : LeftShift;
  else
    Expression;
    PopCmpLess;
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate a Relational "Greater Than" }

procedure TTinyComp.Greater;
begin
  Next;
  case Token of
    '=' : begin
      Next;
      Expression;
      PopCmpGreaterOrEqual;
    end;
    '>' : begin
      RightShift;
    end;
  else
    Expression;
    PopCmpGreater;
  end;
end;


procedure TTinyComp.NEqual;
begin
  Next;
  if Token = '=' then
    NotEqual
  else
    Expected('"!="');
end;

{---------------------------------------------------------------}
{ Parse and Translate a Relation }

procedure TTinyComp.Relation;
begin
  Expression;
  if IsRelop(Token) then begin
    PushPrim;
    case Token of
      '=': Equal;
      '<': Less;
      '>': Greater;
      '!': NEqual;
    end;
//    SetZeroCC;
  end;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Factor with Leading NOT }

procedure TTinyComp.NotFactor;
begin
  if Token = '!' then begin
    Next;
    Relation;
    NotIt;
  end
  else
    Relation;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Term }

procedure TTinyComp.BoolTerm;
begin
  NotFactor;
  while Token = '&' do begin
    PushPrim;
    Next;
    NotFactor;
    PopAnd;
  end;
end;

{--------------------------------------------------------------}
{ Recognize and Translate a Boolean OR }

procedure TTinyComp.BoolOr;
begin
  Next;
  BoolTerm;
  PopOr;
end;

{--------------------------------------------------------------}
{ Recognize and Translate an Exclusive Or }

procedure TTinyComp.BoolXor;
begin
  Next;
  BoolTerm;
  PopXor;
end;

{---------------------------------------------------------------}
{ Parse and Translate a Boolean Expression }

procedure TTinyComp.BoolExpression;
begin
  fCCSet := False;
  BoolTerm;
  while IsOrOp(Token) do begin
    PushPrim;
    case Look of
      '|': BoolOr;
      '~': BoolXor;
    end;
  end;
end;

procedure TTinyComp.StringConst;
var
  LitName : string;
  bIsNull : boolean;
begin
  bIsNull := False;
  if NumConsts >= MAXCONSTS then
    AbortMsg('Constants Table is full');
  LitName := Format('CS%d', [CCount]);
  if IsKnown(LitName) <> 0 then
    Duplicate(LitName);
  Value := '';
  if Look = '"' then
    bIsNull := True
  else
  begin
    repeat
      Value := Value + Look;
      GetChar;
    until (Look = '"') or (Look = #0) or endofallsource;
    CS_Value[NumConsts] := Value;
  end;
  CS_Name[NumConsts] := LitName;
  if bIsNull then begin
    CS_Type[NumConsts] := TOK_BYTEDEF;
    CS_Value[NumConsts] := '0';
  end
  else begin
    CS_Type[NumConsts] := TOK_CHARDEF;
  end;
  Next;
  inc(NumConsts);
  inc(CCount);
  // output assembly ?
  EmitLn('; mov __StrBuf,'+LitName);
  Next;
end;

procedure TTinyComp.DoCall(procname : string);
var
  procpos, protocount, acount, pspace : integer;
  dt : char;
begin
  pspace := 0;
  acount := 0;
  procpos := Locate(procname);
  protocount := GS_ParamCount[procpos];
  Next;
  MatchString('(');
  while Token <> ')' do begin
    if acount = protocount then
      AbortMsg('Too many parameters');
    dt := GS_ParamType[procpos, acount];
    pspace := pspace + SizeOfType(dt);
    if Token = '"' then
      StringConst
    else
      Expression;
    inc(acount);
    Scan;
    if Token = ',' then begin
      Next;
      Scan;
    end;
  end;
  if protocount <> acount then
    AbortMsg('Too few parameters');
  MatchString(')');
  // now output the procedure call assembler
  // set parameter values ???
  EmitLn('call '+procname);
  Scan;
  // pop frame pointer?
  if pspace > 0 then
  begin
    // shrink stack?
  end;
end;


procedure TTinyComp.ArrayAssignment(const name : string);
begin
  Next;
  Expression;
  MatchString(']');
  // do funky stuff for array handling
  EmitLn('; funky array handling code');
  MatchString('=');
  BoolExpression;
  // move RHS to array[idx]
  EmitLn('; move RHS to array[idx]');
end;

procedure TTinyComp.MathAssignment(const name : string);
var
  savedtoken : char;
begin
  // Look has to be '=', '+', or '-' or it's all messed up
  if Look = '=' then
  begin
    savedtoken := Token;
    Next; // move to '='
    Next; // move to next token
    BoolExpression;
    case savedtoken of
      '+' : StoreAdd(name);
      '-' : StoreSub(name);
      '*' : StoreMul(name);
      '/' : StoreDiv(name);
      '%' : StoreMod(name);
      '&' : StoreAnd(name);
      '|' : StoreOr(name);
      '^' : StoreXor(name);
    end;
  end
  else if (Token = '+') and (Look = '+') then
  begin
    Next; // move to second +
    Next;
    Semi;
    StoreInc(name, 1);
  end
  else if (Token = '-') and (Look = '-') then
  begin
    Next; // move to second -
    Next;
    Semi;
    StoreDec(name, 1);
  end
  else if (Token = '+') and (Look = '-') then
  begin
    Next; // move to -
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreSign(name);
    end;
  end
  else if (Token = '|') and (Look = '|') then
  begin
    Next; // move to second |
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreAbs(name);
    end;
  end
  else if ((Token = '>') and (Look = '>')) or ((Token = '<') and (Look = '<')) then
  begin
    savedtoken := Token;
    Next; // move to second > or <
    if Look = '=' then
    begin
      Next; // move to '='
      Next; // move to next token
      BoolExpression;
      StoreShift(savedtoken='>', name);
    end;
  end
  else
    AbortMsg('invalid assignment');
end;

{--------------------------------------------------------------}
{ Parse and Translate an Assignment Statement }

procedure TTinyComp.Assignment;
var
  Name: string;
  dt : char;
begin
  if (IsParam(Value) = 0) and not InTable(Value) and
     (IsLocal(Value) = 0) then
    Undefined(Value);
  Name := Value;
  dt := DataType(Name);
  if dt = TOK_PROCEDURE then begin
    DoCall(Name);
  end
//  else if dt = TOK_APIFUNC then begin
//    DoCallAPIFunc(Name);
//    Store
//  end
  else begin
    Next;
    if Token = '[' then
    begin
      ArrayAssignment(Name);
    end
    else if Token in ['+','-','/','*','%','&','|','^','>','<'] then
    begin
      MathAssignment(Name);
    end
    else
    begin
      MatchString('=');
      BoolExpression;
      Store(Name);
    end;
  end;
end;

{---------------------------------------------------------------}
{ Recognize and Translate an IF Construct }

procedure TTinyComp.DoIf(const lend, lstart : string);
var
  L1, L2: string;
begin
  Next;
  MatchString('(');
  BoolExpression;
  MatchString(')');
  L1 := NewLabel;
  L2 := L1;
  BranchFalse(L1);
  Block(lend, lstart);
  if Token = TOK_ELSE then begin
    Next;
    L2 := NewLabel;
    Branch(L2);
    PostLabel(L1);
    Block(lend, lstart);
  end;
  PostLabel(L2);
end;


{--------------------------------------------------------------}
{ Parse and Translate a WHILE Statement }

procedure TTinyComp.DoWhile;
var
  L1, L2: string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  PostLabel(L1);
  BoolExpression;
  MatchString(')');
  BranchFalse(L2);
  Block(L2, L1);
  Branch(L1);
  PostLabel(L2);
end;

procedure TTinyComp.DoDoWhile;
var
  L1, L2: string;
begin
  Next;
  L1 := NewLabel;
  L2 := NewLabel;
  PostLabel(L1);
  Block(L2, L1);
  MatchString('while');
  MatchString('(');
  BoolExpression;
  MatchString(')');
  BranchFalse(L2);
  Branch(L1);
  PostLabel(L2);
end;

procedure TTinyComp.DoRepeat;
var
  L1, L2: string;
  svar : string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  Expression;
  MatchString(')');
  push;
  svar := tos;
  EmitLn('mov ' + svar + ',__D0');
  PostLabel(L1);
  StoreDec(svar);
  EmitLn('brtst LT,' + L2 + ', ' + svar);
  Block(L2, L1);
  Branch(L1);
  PostLabel(L2);
  pop;
end;

procedure TTinyComp.DoSwitch;
begin

end;

procedure TTinyComp.DoAsm;
var
  asmStr : string;
begin
// gather everything within asm block and output it
  Next;
  MatchString('{');
  asmStr := Value + ' ';
  repeat
    repeat
      GetCharX;
      if Look <> '}' then
        asmStr := asmStr + Look;
    until (Look = LF) or (Look = '}') or endofallsource;
//    if Look = LF then
//      IncLineNumber;
    asmStr := Trim(asmStr);
    if (asmStr <> '') or (Look <> '}') then
      EmitLnNoTab(asmStr);
    asmStr := '';
  until (Look = '}') or endofallsource;
  GetChar; // get the '}'
  Next;
end;

{--------------------------------------------------------------}
{ Parse and Translate a FOR Statement }

procedure TTinyComp.DoFor;
var
  L1, L2, L3, L4: string;
begin
  Next;
  MatchString('(');
  L1 := NewLabel;
  L2 := NewLabel;
  L3 := NewLabel;
  L4 := NewLabel;
  Assignment;
  Semi;
  PostLabel(L1);
  BoolExpression;
  Semi;
  BranchFalse(L2);
  Branch(L3);
  PostLabel(L4);
  Assignment;
  Semi;
  MatchString(')');
  Branch(L1);
  PostLabel(L3);
  Block(L2, L4);
  Branch(L4);
  PostLabel(L2);
end;

{--------------------------------------------------------------}
{ Read a Single Variable }

procedure TTinyComp.ReadVar;
begin
  CheckIdent;
  CheckTable(Value);
  ReadIt(Value);
  Next;
end;


{--------------------------------------------------------------}
{ Process a Read Statement }

procedure TTinyComp.DoRead;
begin
  Next;
  MatchString('(');
  ReadVar;
  while Token = ',' do begin
    Next;
    ReadVar;
  end;
  MatchString(')');
end;


{--------------------------------------------------------------}
{ Process a Write Statement }

procedure TTinyComp.DoWrite;
begin
  Next;
  MatchString('(');
  Expression;
  WriteIt;
  while Token = ',' do begin
    Next;
    Expression;
    WriteIt;
  end;
  MatchString(')');
end;

{--------------------------------------------------------------}
{ Process a wait Statement }

procedure TTinyComp.DoWait;
begin
  Next;
  MatchString('(');
  Expression;
  MatchString(')');
  EmitLn('waitv __D0');
end;

{--------------------------------------------------------------}
{ Process a cls Statement }

procedure TTinyComp.DoCls;
begin
  Next;
  MatchString('(');
  MatchString(')');
  EmitLn('set __clsSSMArgs.ScreenMode, 0');
  EmitLn('syscall SetScreenMode, __clsSSMArgs');
end;

function IndexOfAPICommand(const name : string) : integer;
begin
  for Result := Low(APIList) to High(APIList) do
  begin
    if APIList[Result] = name then
      Exit;
  end;
  Result := -1;
end;

procedure TTinyComp.DoAPICommands(const lend, lstart : string);
var
  idx : integer;
begin
  idx := IndexOfAPICommand(Value);
  case idx of
    API_BREAK    : DoBreakContinue(idx, lend);
    API_CONTINUE : DoBreakContinue(idx, lstart);
    API_RETURN   : DoReturn;
    API_READ     : DoRead;
    API_WRITE    : DoWrite;
    API_WAIT     : DoWait;
    API_CLS      : DoCls;
    API_ONFWD,
    API_ONREV    : DoOnFwdRev;
    API_ONFWDREG,
    API_ONREVREG : DoOnFwdRevReg;
    API_OFF,
    API_COAST    : DoStopMotors;
    API_ONFWDSYNC,
    API_ONREVSYNC : DoOnFwdRevSync;
    API_ROTATEMOTOR,
    API_ROTATEMOTOREX : DoRotateMotors(idx);
    API_SETSENSORTYPE,
    API_SETSENSORMODE : DoSetSensorTypeMode(idx);
    API_CLEARSENSOR,
    API_SETSENSORTOUCH,
    API_SETSENSORLIGHT,
    API_SETSENSORSOUND,
    API_SETSENSORULTRASOUND,
    API_RESETSENSOR : DoClearSetResetSensor;
    API_READSENSOR,
    API_READSENSORUS : DoReadSensor;
    API_TEXTOUT,
    API_NUMOUT : DoTextNumOut(idx);
    API_PLAYTONE : DoPlayTone;
    API_PLAYFILE : DoPlayFile;
    API_RESETSLEEP : DoResetSleepTimer;
    API_PRECEDES,
    API_FOLLOWS : DoPrecedesFollows;
    API_ACQUIRE,
    API_RELEASE : DoAcquireRelease;
  else
    AbortMsg('Unknown API command');
  end;
end;

{--------------------------------------------------------------}
{ Parse and Translate a Single Statement }

procedure TTinyComp.Statement(const lend, lstart : string);
begin
  ProcessDirectives;
  case Token of
    TOK_IF:         DoIf(lend, lstart);
    TOK_WHILE:      DoWhile;
    TOK_FOR:        DoFor;
    TOK_DO:         DoDoWhile;
    TOK_REPEAT:     DoRepeat;
    TOK_SWITCH:     DoSwitch;
    TOK_ASM:        DoAsm;
    TOK_API:        DoAPICommands(lend, lstart);
    TOK_IDENTIFIER: Assignment;
  end;
end;


{--------------------------------------------------------------}
{ Parse and Translate a Block of Statements }

procedure TTinyComp.Block(const lend, lstart : string);
var
  bBlockStatement : boolean;
begin
  bBlockStatement := Value = '{';
  if bBlockStatement then
  begin
    Next;
    BlockStatements(lend, lstart);
    MatchString('}');
  end
  else
  begin
    Scan;
    Statement(lend, lstart);
    Semi;
    Scan;
  end;
end;

procedure TTinyComp.BlockStatements(const lend, lstart: string);
begin
  Scan;
  while not (Token in [TOK_END, TOK_ELSE]) do begin
    Statement(lend, lstart);
    Semi;
    Scan;
  end;
end;

{--------------------------------------------------------------}
{ Allocate Storage for a Variable }

procedure TTinyComp.Alloc(dt : char);
var
  savedval : string;
  ival : string;
begin
  Next;
  if Token <> TOK_IDENTIFIER then Expected('Variable Name');
  CheckDup(Value);
  AddEntry(Value, dt);
  // optional initial value
  savedval := Value;
  ival := '';
  Next;
  if (Token = ',') or (Token = ';') then
  begin
    Allocate(savedval, ival, dt);
  end
  else if Token = '=' then
  begin
    Next;
    ival := Value;
    if Token = '-' then
    begin
      Next;
      ival := ival + Value;
    end;
    Allocate(savedval, ival, dt);
    Next;
  end
  else
    Next;
end;


function TTinyComp.GetVariableType(vt : char; bUnsigned : boolean) : char;
begin
  if not bUnsigned then
    Result := vt
  else
    case vt of
      TOK_LONGDEF : Result := TOK_ULONGDEF;
      TOK_SHORTDEF : Result := TOK_USHORTDEF;
      TOK_CHARDEF : Result := TOK_BYTEDEF;
    else
      Result := vt;
    end;
end;

{--------------------------------------------------------------}
{ Parse and Translate Global Declarations }

procedure TTinyComp.TopDecls;
var
  vt : char;
  bUnsigned : boolean;
begin
  bUnsigned := False;
  Scan;
  while not (Token in [TOK_TASK, TOK_PROCEDURE]) and not endofallsource do
  begin
    case Token of
      TOK_DIRECTIVE : begin
        ProcessDirectives;
        Scan;
      end;
      TOK_UNSIGNED : begin
        Next;
        Scan;
        bUnsigned := True;
      end;
      TOK_LONGDEF, TOK_SHORTDEF,
      TOK_CHARDEF, TOK_BYTEDEF, TOK_MUTEXDEF : begin
        vt := Token;
        Alloc(GetVariableType(vt, bUnsigned));
        while Token = ',' do
          Alloc(GetVariableType(vt, bUnsigned));
        Semi;
        Scan;
        bUnsigned := False;
      end;
    else
      // nothing here right now
      Semi;
      Scan;
    end;
  end;
end;

procedure TTinyComp.AddLocal(name : string; dt : char);
begin
  if (IsParam(name) <> 0) or (IsLocal(name) <> 0) then
    Duplicate(name);
  LS_Name[NumLocals] := name;
  inc(NumLocals);
end;

procedure TTinyComp.AllocLocal(const sub : string; dt : char);
var
  Name : string;
begin
  Next;
  if Token <> TOK_IDENTIFIER then
    Expected('Variable Name');
  AddLocal(sub+'_'+Value, dt);
  Name := Value;
  Next;
  Allocate(sub+'_'+Name, Value, dt);
end;

procedure TTinyComp.DoLocals(const sub : string);
var
  bIsUnsigned : boolean;
  dt : char;
begin
  bIsUnsigned := False;
  ClearLocals;
  Scan;
  while (Token in [TOK_UNSIGNED, TOK_LONGDEF, TOK_SHORTDEF,
    TOK_CHARDEF, TOK_BYTEDEF, TOK_MUTEXDEF]) and not endofallsource do
  begin
    case Token of
      TOK_UNSIGNED : begin
        Next;
        Scan;
        bIsUnsigned := True;
      end;
      TOK_LONGDEF, TOK_SHORTDEF,
      TOK_CHARDEF, TOK_BYTEDEF, TOK_MUTEXDEF : begin
        dt := Token;
        AllocLocal(sub, GetVariableType(dt, bIsUnsigned));
        while Token = ',' do
          AllocLocal(sub, GetVariableType(dt, bIsUnsigned));
        Semi;
        Scan;
        bIsUnsigned := False;
      end;
    else
      Expected('Valid Program Block Statement');
    end;
  end;
end;

function TTinyComp.FormalList(protoexists : boolean; procname : string) : integer;
var
  protocount : integer;
  pltype : integer;
  pcount, procpos : integer;
  ptype : char;
  varnam : string;
  bIsUnsigned : boolean;
begin
  protocount := 0;
  pcount := 0;
  pltype := 0;
  ptype := #0;
  procpos := Locate(procname);
  if protoexists then
    protocount := GS_ParamCount[procpos];
  while Token <> ')' do begin
    Scan;
    if Token = TOK_UNSIGNED then begin
      bIsUnsigned := True;
      Next;
      Scan;
      pltype := 1;
    end
    else
      bIsUnsigned := False;
    if Token in [TOK_CHARDEF, TOK_BYTEDEF, TOK_SHORTDEF, TOK_LONGDEF, TOK_MUTEXDEF] then
    begin
      if protoexists then
        Expected('Parameter List');
      ptype := Token;
      pltype := 1;
      Next;
      Scan;
      if (Token <> ',') and (Token <> ')') and (Token <> TOK_IDENTIFIER) then
        AbortMsg('Unexpected character encountered');
    end
    else
      if bIsUnsigned then
        AbortMsg('Missing Data Type');
    if Token = TOK_IDENTIFIER then
    begin
      varnam := Value;
      Next;
      Scan;
      if pltype = 1 then
        pltype := 3
      else
        pltype := 2;
    end;
    ptype := GetVariableType(ptype, bIsUnsigned);
    case pltype of
      1 : begin
        GS_ParamType[procpos, protocount] := ptype;
        if protocount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        inc(protocount);
      end;
      2 : begin
        if not protoexists then
          Expected('Data type');
        if pcount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        inc(pcount);
        if pcount > protocount then
          AbortMsg(sTooManyArgs);
        AddParam(varnam, GS_ParamType[procpos, pcount]);
      end;
      3 : begin
        if protoexists then
          AbortMsg('Data types already defined in prototype');
        GS_ParamType[procpos, protocount] := ptype;
        if protocount >= MAXPARAMS then
          AbortMsg(sMaxParamCountExceeded);
        inc(protocount);
        inc(pcount);
        AddParam(varnam, ptype);
      end;
    end;
    while Token = ',' do begin
      Next;
      Scan;
      if (pltype = 1) or (pltype = 3) then begin
        if Token = TOK_UNSIGNED then begin
          bIsUnsigned := True;
          Next;
          Scan;
        end
        else
          bIsUnsigned := False;
        if Token in [TOK_CHARDEF, TOK_BYTEDEF, TOK_SHORTDEF, TOK_LONGDEF, TOK_MUTEXDEF] then
        begin
          ptype := Token;
          Next;
          Scan;
        end
        else if bIsUnsigned then
          AbortMsg('Missing Data Type');
      end;
      if (pltype = 2) or (pltype = 3) then begin
        if Token = TOK_IDENTIFIER then begin
          varnam := Value;
          Next;
          Scan;
        end
        else
          Expected(sVariableName);
      end;
      case pltype of
        1 : begin
          GS_ParamType[procpos, protocount] := ptype;
          if protocount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          inc(protocount);
        end;
        2 : begin
          if pcount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          inc(pcount);
          if pcount > protocount then
            AbortMsg(sTooManyArgs);
          AddParam(varnam, GS_ParamType[procpos, pcount]);
        end;
        3 : begin
          GS_ParamType[procpos, protocount] := ptype;
          if protocount >= MAXPARAMS then
            AbortMsg(sMaxParamCountExceeded);
          inc(protocount);
          inc(pcount);
          AddParam(varnam, ptype);
        end;
      end;
    end; // while Token = ','
  end; // while Token <> ')'
  if protoexists and (pcount <> protocount) then
    AbortMsg('Too few arguments');
  case pltype of
    0 : begin
      // nothing
    end;
    1 : begin
      GS_ParamCount[procpos] := protocount;
    end;
    2 : begin
      GS_ParamCount[procpos] := pcount;
    end;
    3 : begin
      GS_ParamCount[procpos] := pcount;
    end;
  end;
  Result := pltype;
end;

procedure TTinyComp.ProcedureBlock;
var
  Name : string;
  procexists : integer;
  protoexists, bIsSub : boolean;
  pltype : integer;
begin
  while Token in [TOK_PROCEDURE, TOK_TASK] do
  begin
    bIsSub := Token = TOK_PROCEDURE;
    Next;
    if Token <> TOK_IDENTIFIER then
      Expected('Procedure Name');
    Name := Value;
    procexists := Locate(Name);
    protoexists := False;
    if procexists <> 0 then begin
      if GS_Type[procexists] <> TOK_PROCEDURE then
        Duplicate(Name);
      if GS_Size[procexists] = 0 then
        protoexists := True
      else
        Duplicate(Name);
    end
    else begin
      AddEntry(Name, TOK_PROCEDURE);
    end;
    Next;
    MatchString('(');
    pltype := FormalList(protoexists, Name);
    MatchString(')');
    Semi;
    Scan;
    if Token = TOK_BEGIN then
    begin
      if pltype = 1 then
        AbortMsg('Not valid for a prototype');
      if protoexists then
        GS_Size[procexists] := 1
      else
        GS_Size[NumGlobals] := 1;
      Prolog(Name, bIsSub);
      MatchString('{');
      DoLocals(Name);
      BlockStatements();
      MatchString('}');
      Epilog(bIsSub);
      Scan;
    end
    else
    begin
      if protoexists then
        Expected('Prototype already defined - "begin"');
      Scan;
    end;
    ClearParams;
    TopDecls;
  end;
end;

{--------------------------------------------------------------}
{ Initialize }

procedure TTinyComp.Init;
var
  i : integer;
begin
  fCurrentLine := '';
  totallines := 1;
  for i := 1 to MAXINCL do
    linenumber[i] := 1;
  ClearParams;
  sd := 0;
  maxsd := 0;
  GetChar;
  Next;
end;

{--------------------------------------------------------------}
{  Parse and Translate a Program }

procedure TTinyComp.Prog;
begin
  Header;
  TopDecls;
  if Token in [TOK_PROCEDURE, TOK_TASK] then
    ProcedureBlock;
  Trailer;
end;

{ TTinyComp }

constructor TTinyComp.Create;
begin
  fNBCSrc := TStringList.Create;
  fMS := TMemoryStream.Create;
  fMessages := TStringList.Create;
  fIncludeDirs := TStringList.Create;
  fAPIFunctions := TStringList.Create;
  TStringList(fAPIFunctions).Sorted := True;
  LoadAPIFunctions;
end;

destructor TTinyComp.Destroy;
begin
  FreeAndNil(fNBCSrc);
  FreeAndNil(fMS);
  FreeAndNil(fMessages);
  FreeAndNil(fIncludeDirs);
  FreeAndNil(fAPIFunctions);
  inherited;
end;

procedure TTinyComp.InternalParseStream;
begin
  try
    fBadProgram     := False;
    fProgErrorCount := 0;
    PreProcess;
    fMS.Position := 0;
    Init;
    Prog;
  except
    on E : Exception do
    begin
      fBadProgram := True;
      ReportProblem(linenumber[slevel], CurrentFile, sException, E.Message, true);
    end;
  end;
end;

procedure TTinyComp.Parse(aStrings: TStrings);
begin
  Clear;
  aStrings.SaveToStream(fMS);
  InternalParseStream;
end;

procedure TTinyComp.Parse(aStream: TStream);
begin
  Clear;
  fMS.CopyFrom(aStream, 0);
  InternalParseStream;
end;

procedure TTinyComp.Parse(const aFilename: string);
begin
  Clear;
  fMS.LoadFromFile(aFilename);
  InternalParseStream;
end;

procedure TTinyComp.Clear;
begin
  fNBCSrc.Clear;
  fMessages.Clear;
  fTempChar := ' ';
end;

{--------------------------------------------------------------}
{ Recognize and Translate a break/continue }

procedure TTinyComp.DoBreakContinue(idx : integer; const lbl: string);
var
  val : string;
begin
  val := APIList[idx];
  MatchString(val);
  Semi;
  if lbl <> '' then
    Branch(lbl)
  else
    AbortMsg(Format('%s is invalid outside of a loop', [val]));
end;

procedure TTinyComp.DoOnFwdRev;
var
  op, arg1 : string;
begin
  op := Value;
  Next;
  MatchString('(');
  arg1 := Value;
  Next;
  MatchString(',');
  Expression;
  MatchString(')');
  EmitLn(op + '(' + arg1 + ', __D0)');
end;

procedure TTinyComp.DoOnFwdRevReg;
var
  op, arg1, svar : string;
begin
  op := Value;
  Next;
  MatchString('(');
  arg1 := Value;
  Next;
  MatchString(',');
  Expression;
  MatchString(',');
  push;
  svar := tos;
  EmitLn('mov ' + svar + ',__D0');
  Expression;
  MatchString(')');
  EmitLn(op + '(' + arg1 + ', ' + svar + ', __D0)');
  pop;
end;

procedure TTinyComp.DoOnFwdRevSync;
var
  op, arg1, svar : string;
begin
  //OnFwdSync(ports, pwr, turnpct)
  op := Value;
  Next;
  MatchString('(');
  arg1 := Value;
  Next;
  MatchString(',');
  Expression;
  MatchString(',');
  push;
  svar := tos;
  EmitLn('mov ' + svar + ', __D0');
  Expression;
  MatchString(')');
  EmitLn(op + '(' + arg1 + ', ' + svar + ', __D0)');
  pop;
end;

procedure TTinyComp.DoRotateMotors(idx: integer);
var
  op, ports, pwr, angle, turnpct : string;
begin
 //RotateMotor(ports, pwr, angle)
 //RotateMotorEx(ports, pwr, angle, turnpct, bSync)
  op := Value;
  Next;
  MatchString('(');
  ports := Value;
  Next;
  MatchString(',');
  Expression;
  MatchString(',');
  push;
  pwr := tos;
  EmitLn('mov ' + pwr + ', __D0');
  Expression;
  if idx = API_ROTATEMOTOR then
  begin
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s)', [op, ports, pwr, '__D0']))
  end
  else
  begin
    MatchString(',');
    push;
    angle := tos;
    EmitLn('mov ' + angle + ', __D0');
    Expression;
    MatchString(',');
    push;
    turnpct := tos;
    EmitLn('mov ' + turnpct + ', __D0');
    BoolExpression;
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s, %s, %s)', [op, ports, pwr, angle, turnpct, '__D0']));
    pop;
    pop;
  end;
  pop;
end;

procedure TTinyComp.DoClearSetResetSensor;
var
  op, port : string;
begin
  //ClearSensor(port)
  //SetSensorTouch(port)
  //SetSensorLight(port)
  //SetSensorSound(port)
  //SetSensorUltrasonic(port)
  //ResetSensor(port)
  op := Value;
  Next;
  MatchString('(');
  port := Value;
  Next;
  MatchString(')');
  EmitLn(op + '(' + port + ')');
end;

procedure TTinyComp.DoSetSensorTypeMode(idx : integer);
var
  port : string;
begin
  //SetSensorType(port,type)
  //SetSensorMode(port,mode)
  Next;
  MatchString('(');
  port := Value;
  Next;
  MatchString(',');
  Expression;
  MatchString(')');
  if idx = API_SETSENSORTYPE then
    EmitLn('setin __D0, ' + port + ', Type')
  else
    EmitLn('setin __D0, ' + port + ', InputMode');
end;

procedure TTinyComp.DoReadSensor;
var
  op, port, val : string;
begin
  //ReadSensor(port,value)
  //ReadSensorUS(port,value)
  op := Value;
  Next;
  MatchString('(');
  port := Value;
  Next;
  MatchString(',');
  val := Value;
  CheckIdent;
  CheckTable(Value);
  Next;
  MatchString(')');
  EmitLn(Format('%s(%s, %s)', [op, port, val]));
end;

procedure TTinyComp.DoTextNumOut(idx: integer);
var
  op, x, y, cls, val : string;
begin
  //TextOut(x,y,cls,txt)
  //NumOut(x,y,cls,num)
  op := Value;
  Next;
  MatchString('(');
  Expression;
  MatchString(',');
  push;
  x := tos;
  EmitLn('mov ' + x + ', __D0');
  Expression;
  MatchString(',');
  push;
  y := tos;
  EmitLn('mov ' + y + ', __D0');
  BoolExpression;
  MatchString(',');
  push;
  cls := tos;
  EmitLn('mov ' + cls + ', __D0');
  if idx = API_NUMOUT then
  begin
    Expression;
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s, __D0)', [op, x, y, cls]));
  end
  else
  begin
    val := Value;
    Next;
    MatchString(')');
    EmitLn(Format('%s(%s, %s, %s, %s)', [op, x, y, cls, val]));
  end;
  pop;
  pop;
  pop;
end;

procedure TTinyComp.DoPlayTone;
var
  freq : string;
begin
  //PlayTone(freq,dur)
  Next;
  MatchString('(');
  Expression;
  MatchString(',');
  push;
  freq := tos;
  EmitLn('mov ' + freq + ', __D0');
  Expression;
  MatchString(')');
  EmitLn('PlayTone(' + freq + ', __D0)');
  pop;
end;

procedure TTinyComp.DoPlayFile;
var
  arg1 : string;
begin
  //PlayFile(file)
  Next;
  MatchString('(');
  arg1 := Value;
  Next;
  MatchString(')');
  EmitLn('PlayFile(' + arg1 + ')');
end;

procedure TTinyComp.DoAcquireRelease;
var
  op, val : string;
begin
 // Acquire(mutex);
 // Release(mutex);
  op := LowerCase(Value);
  Next;
  MatchString('(');
  val := Value;
  CheckIdent;
  CheckTable(val);
  if DataType(val) <> TOK_MUTEXDEF then Expected('Mutex');
  Next;
  MatchString(')');
  EmitLn(op + ' ' + val);
end;

procedure TTinyComp.DoPrecedesFollows;
var
  op, val : string;
begin
  // Precedes(x, y, z, ...);
  // Follows(x, y, z);
  op := LowerCase(Value);
  Next;
  MatchString('(');
  val := Value;
  CheckIdent;
  CheckTable(val);
  Next;
  while Value = ',' do
  begin
    val := val + Value; // the comma
    Next;
    CheckIdent;
    CheckTable(Value);
    val := val + ' ' + Value; // the next value
    Next;
  end;
  MatchString(')');
  EmitLn(op + ' ' + val);
end;

procedure TTinyComp.DoResetSleepTimer;
begin
  Next;
  MatchString('(');
  MatchString(')');
  EmitLn('syscall KeepAlive, __KeepAliveArgs');
end;

procedure TTinyComp.DoReturn;
begin
  MatchString(APIList[API_RETURN]);
  Semi;
  EmitLn('return');
end;

procedure TTinyComp.DoStopMotors;
var
  op, arg1 : string;
begin
  op := Value;
  Next;
  MatchString('(');
  arg1 := Value;
  Next;
  MatchString(')');
  EmitLn(op + '(' + arg1 + ')');
end;

procedure TTinyComp.PreProcess;
var
  P : TLangPreprocessor;
begin
  P := TLangPreprocessor.Create(TNQCLexer, ExtractFilePath(ParamStr(0)));
  try
    P.AddIncludeDirs(IncludeDirs);
    P.Preprocess(CurrentFile, fMS);
  finally
    P.Free;
  end;
end;

function Replace(const str : string; const src, rep : string) : string;
begin
{$IFDEF FAST_MM}
  Result := FastReplace(str, src, rep);
{$ELSE}
  Result := StringReplace(str, src, rep, [rfReplaceAll]);
{$ENDIF}
end;

procedure TTinyComp.ProcessDirectives;
begin
  while Token = TOK_DIRECTIVE do
  begin
    Next;
    // look for #line statements
    if (Token = TOK_IDENTIFIER) and (LowerCase(Value) = 'line') then
    begin
      SkipDirectiveLine;
      HandlePoundLine;
      Next;
    end
    else
    begin
      SkipDirectiveLine;
      Next;
    end;
    Scan;
  end;
end;

procedure TTinyComp.HandlePoundLine;
var
  i : integer;
  tmpLine, tmpFile : string;
begin
  i := Pos('#line ', fDirLine);
  if i = 1 then
  begin
    // this is a special preprocessor line
    tmpLine := Trim(fDirLine);
    Delete(tmpLine, 1, 6);
    i := Pos(' ', tmpLine);
    linenumber[slevel] := StrToIntDef(Copy(tmpLine, 1, i - 1), linenumber[slevel]);
    IncLineNumber;
    Delete(tmpLine, 1, i);
    tmpFile      := Replace(tmpLine, '"', '');
    CurrentFile  := tmpFile;
  end;
end;

procedure TTinyComp.IncLineNumber;
begin
  linenumber[slevel] := linenumber[slevel] + 1;
  inc(totallines);
end;

procedure TTinyComp.DoCallAPIFunc(procname: string);
var
  arg : string;
begin
  MatchString('(');
  if procname = 'Random' then
  begin
    BoolExpression;
    MatchString(')');
    EmitLn('Random(__D0, __D0)');
  end
  else if procname = 'abs' then
  begin
    BoolExpression;
    MatchString(')');
    EmitLn('abs __D0, __D0');
  end
  else if procname = 'sign' then
  begin
    BoolExpression;
    MatchString(')');
    EmitLn('sign __D0, __D0');
  end
  else if procname = 'InputFieldValue' then
  begin
    BoolExpression;
    MatchString(',');
    arg := Value;
    MatchString(')');
    EmitLn('getin __D0, __D0, ' + arg);
  end
  else if procname = 'OutputFieldValue' then
  begin
    BoolExpression;
    MatchString(',');
    arg := Value;
    MatchString(')');
    EmitLn('getout __D0, __D0, ' + arg);
  end
  else if procname = 'IOMA' then
  begin
    arg := Value;
    CheckNumeric;
    MatchString(')');
    EmitLn('mov __D0, ' + arg);
  end
  else if procname = 'GetTick' then
  begin
    MatchString(')');
    EmitLn('gettick __D0');
  end
end;

procedure TTinyComp.LoadAPIFunctions;
begin
  fAPIFunctions.Add('abs');
  fAPIFunctions.Add('sign');
  fAPIFunctions.Add('Random');
  fAPIFunctions.Add('InputFieldValue');
  fAPIFunctions.Add('OutputFieldValue');
  fAPIFunctions.Add('IOMA');
  fAPIFunctions.Add('GetTick');
end;

(*
FileOpenRead (0), FileOpenWrite (1), FileOpenAppend (2), FileRead (3),
FileWrite (4), FileClose (5), FileResolveHandle (6), FileRename (7),
FileDelete (8), SoundGetState (11),
SoundSetState (12), DrawText (13), DrawPoint (14), DrawLine (15),
DrawCircle (16), DrawRect (17), DrawGraphic (18), SetScreenMode (19),
ReadButton (20), CommLSWrite (21), CommLSRead (22), CommLSCheckStatus (23),
GetStartTick (25), MessageWrite (26), MessageRead (27),
CommBTCheckStatus (28), CommBTWrite (29), KeepAlive (31), IOMapRead (32),
IOMapWrite (33)

*)

end.
